# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from spox._shape import Shape


@pytest.fixture(
    params=[
        [(2, 3, 4, 5), (), (2, 3, 4, 5)],
        [(2, 3, 4, 5), (5,), (2, 3, 4, 5)],
        [(2, 3, 4, 5), (2, 1, 1, 5), (2, 3, 4, 5)],
        [(2, 3, 4, 5), (1, 3, 1, 5), (2, 3, 4, 5)],
        [("x", "y", None), ("x", "y", None), ("x", "y", None)],
        [("x", "y", None), (1, "y", 1), ("x", "y", None)],
        [("x", "y", None), (1, None), ("x", "y", None)],
        [(1, 2, 3), (None, None, None), (None, 2, 3)],
        [("x", None), (3, 2, "x", 7), (3, 2, "x", 7)],
        [("x", None), ("y", 2, "x", 7), ("y", 2, "x", 7)],
        [("x", "y", None), ("x", "z", None), ("x", None, None)],
        [("x", "y", None), ("x", "z", "z"), ("x", None, None)],
        [("x", "y", None), (1, "z", 1), ("x", None, None)],
        [("x", "y", None), (1, "z", 1), ("x", None, None)],
        [(5, 7), ("x", None), (5, 7)],
        [(1, 2, 3), ("x", "y", "z"), ("x", 2, 3)],
        [(1, 2, 3), (1, 2, "n"), (1, 2, 3)],
    ]
)
def broadcast_shapes(request):
    return tuple(Shape.from_simple(sh) for sh in request.param)


@pytest.fixture(
    params=[
        [(2, 3, 4, 5), (6,)],
        [(2, 3, 4, 5), (6,)],
        [(2, 3, 4, 5), (3, 1, 1, 5)],
        [(2, 3, 4, 5), (1, 2, 1, 5)],
        [(2, "x", "y"), (3, "y", "z")],
    ]
)
def no_broadcast_shapes(request):
    return tuple(Shape.from_simple(sh) for sh in request.param)


def test_can_broadcast_true(broadcast_shapes: tuple[Shape, Shape, Shape]):
    first, second, _ = broadcast_shapes
    assert first.can_broadcast(second)


def test_can_broadcast_false(no_broadcast_shapes: tuple[Shape, Shape]):
    first, second = no_broadcast_shapes
    assert not first.can_broadcast(second)


def test_broadcast(broadcast_shapes: tuple[Shape, Shape, Shape]):
    first, second, result = broadcast_shapes
    assert first.broadcast(second) == result



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

import datetime
import importlib
import inspect
import os
import subprocess
import sys
from subprocess import CalledProcessError
from typing import Any, Optional, cast

_mod = importlib.import_module("spox")


project = "spox"
copyright = f"{datetime.date.today().year}, QuantCo, Inc"
author = "QuantCo, Inc."

extensions = [
    "numpydoc",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "nbsphinx",
    "sphinx.ext.doctest",
    "sphinx.ext.linkcode",
    "sphinxcontrib.apidoc",
]

apidoc_module_dir = "../src/spox"
apidoc_output_dir = "api"
apidoc_separate_modules = True
apidoc_extra_args = ["--implicit-namespaces"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]


# Copied and adapted from
# https://github.com/pandas-dev/pandas/blob/4a14d064187367cacab3ff4652a12a0e45d0711b/doc/source/conf.py#L613-L659
# Required configuration function to use sphinx.ext.linkcode
def linkcode_resolve(domain: str, info: dict[str, Any]) -> Optional[str]:
    """Determine the URL corresponding to a given Python object."""
    if domain != "py":
        return None

    module_name = info["module"]
    full_name = info["fullname"]

    _submodule = sys.modules.get(module_name)
    if _submodule is None:
        return None

    _object = _submodule
    for _part in full_name.split("."):
        try:
            _object = getattr(_object, _part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(_object))  # type: ignore
    except TypeError:
        fn = None
    if not fn:
        return None

    try:
        source, line_number = inspect.getsourcelines(_object)
    except OSError:
        line_number = None

    if line_number:
        linespec = f"#L{line_number}-L{line_number + len(source) - 1}"
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(cast(str, _mod.__file__)))

    try:
        # See https://stackoverflow.com/a/21901260
        commit = (
            subprocess.check_output(["git", "rev-parse", "HEAD"])
            .decode("ascii")
            .strip()
        )
    except CalledProcessError:
        # If subprocess returns non-zero exit status
        commit = "main"

    return (
        "https://github.com/quantco/spox"
        f"/blob/{commit}/src/{_mod.__name__.replace('.', '/')}/{fn}{linespec}"
    )



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._type_system import Tensor


def test_normalizer_inference():
    (x,) = arguments(x=Tensor(np.float64, ("N", 5)))
    y = op_ml.normalizer(x)
    assert y.type == Tensor(np.float64, ("N", 5))



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import enum
import warnings
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import Field, dataclass
from typing import Optional, get_type_hints

from . import _type_system
from ._attributes import Attr
from ._exceptions import InferenceWarning
from ._value_prop import PropDict, PropValue
from ._var import Var, _VarInfo


@dataclass
class BaseFields:
    pass


@dataclass
class BaseAttributes(BaseFields):
    def get_fields(self) -> dict[str, None | Attr]:
        """Return a mapping of all fields stored in this object by name."""
        return self.__dict__.copy()


class VarFieldKind(enum.Enum):
    """An enumeration of possible kinds of ONNX input/outputs fields."""

    SINGLE = 0
    OPTIONAL = 1
    VARIADIC = 2


class BaseVars:
    """A collection of `Var`-s used to carry around inputs/outputs of nodes"""

    vars: dict[str, Var | None | Sequence[Var]]

    def __init__(self, vars: dict[str, Var | None | Sequence[Var]]):
        self.vars = vars

    def _unpack_to_any(self) -> tuple[Var | None | Sequence[Var], ...]:
        """Unpack the stored fields into a tuple of appropriate length, typed as Any."""
        return tuple(self.vars.values())

    def _flatten(self) -> Iterator[tuple[str, Var | None]]:
        """Iterate over the pairs of names and values of fields in this object."""
        for key, value in self.vars.items():
            if value is None or isinstance(value, Var):
                yield key, value
            else:
                yield from ((f"{key}_{i}", v) for i, v in enumerate(value))

    def flatten_vars(self) -> dict[str, Var]:
        """Return a flat mapping by name of all the VarInfos in this object."""
        return {key: var for key, var in self._flatten() if var is not None}

    def __getattr__(self, attr: str) -> Var | None | Sequence[Var]:
        """Retrieves the attribute if present in the stored variables."""
        try:
            return self.vars[attr]
        except KeyError:
            raise AttributeError(
                f"{self.__class__.__name__!r} object has no attribute {attr!r}"
            )

    def __setattr__(self, attr: str, value: Var | None | Sequence[Var]) -> None:
        """Sets the attribute to a value if the attribute is present in the stored variables."""
        if attr == "vars":
            super().__setattr__(attr, value)
        else:
            self.vars[attr] = value


@dataclass
class BaseVarInfos(BaseFields):
    def __post_init__(self) -> None:
        # Check if passed fields are of the appropriate types based on field kinds
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            field_type = self._get_field_type(field)
            if field_type == VarFieldKind.SINGLE:
                if not isinstance(value, _VarInfo):
                    raise TypeError(f"Field expected VarInfo, got: {type(value)}.")
            elif field_type == VarFieldKind.OPTIONAL:
                if value is not None and not isinstance(value, _VarInfo):
                    raise TypeError(
                        f"Optional must be VarInfo or None, got: {type(value)}."
                    )
            elif field_type == VarFieldKind.VARIADIC:
                if not isinstance(value, Iterable):
                    raise TypeError(
                        f"Variadic field must be iterable, got '{type(value)}'."
                    )
                # Cast to tuple to avoid accidental mutation
                setattr(self, field.name, tuple(value))
                if bad := {type(var) for var in value} - {_VarInfo}:
                    raise TypeError(
                        f"Variadic field must only consist of VarInfos, got: {bad}."
                    )

    @classmethod
    def _get_field_type(cls, field: Field) -> VarFieldKind:
        """Access the kind of the field (single, optional, variadic) based on its type annotation."""
        # The field.type may be unannotated as per
        # from __future__ import annotations
        field_type = get_type_hints(cls)[field.name]
        if field_type == _VarInfo:
            return VarFieldKind.SINGLE
        elif field_type == Optional[_VarInfo]:
            return VarFieldKind.OPTIONAL
        elif field_type == Sequence[_VarInfo]:
            return VarFieldKind.VARIADIC
        raise ValueError(f"Bad field type: '{field.type}'.")

    def _flatten(self) -> Iterable[tuple[str, _VarInfo | None]]:
        """Iterate over the pairs of names and values of fields in this object."""
        for key, value in self.__dict__.items():
            if value is None or isinstance(value, _VarInfo):
                yield key, value
            else:
                yield from ((f"{key}_{i}", v) for i, v in enumerate(value))

    def __iter__(self) -> Iterator[_VarInfo | None]:
        """Iterate over the values of fields in this object."""
        yield from (v for _, v in self._flatten())

    def __len__(self) -> int:
        """Count the number of fields in this object (should be same as declared in the class)."""
        return sum(1 for _ in self)

    def get_var_infos(self) -> dict[str, _VarInfo]:
        """Return a flat mapping by name of all the VarInfos in this object."""
        return {key: var for key, var in self._flatten() if var is not None}

    def get_fields(self) -> dict[str, None | _VarInfo | Sequence[_VarInfo]]:
        """Return a mapping of all fields stored in this object by name."""
        return self.__dict__.copy()

    @property
    def fully_typed(self) -> bool:
        """Check if all stored variables have a concrete type."""
        return all(
            var.type is not None and var.type._is_concrete
            for var in self.get_var_infos().values()
        )

    def into_vars(self, prop_values: PropDict) -> BaseVars:
        """Populate a `BaseVars` object with the propagated values and this object's var_infos"""

        def _create_var(key: str, var_info: _VarInfo) -> Var:
            ret = Var(var_info, None)

            if var_info.type is None or key not in prop_values:
                return ret

            if (
                not isinstance(var_info.type, _type_system.Optional)
                and prop_values[key] is None
            ):
                return ret

            prop = PropValue(var_info.type, prop_values[key])
            if prop.check():
                ret._value = prop
            else:
                warnings.warn(
                    InferenceWarning(
                        f"Propagated value {prop} does not type-check, dropping. "
                        f"Hint: this indicates a bug with the current value prop backend or type inference."
                    )
                )

            return ret

        ret_dict: dict[str, Var | None | Sequence[Var]] = {}

        for key, var_info in self.__dict__.items():
            if isinstance(var_info, _VarInfo):
                ret_dict[key] = _create_var(key, var_info)
            else:
                ret_dict[key] = [
                    _create_var(f"{key}_{i}", v) for i, v in enumerate(var_info)
                ]

        return BaseVars(ret_dict)


@dataclass
class BaseInputs(BaseVarInfos):
    pass


@dataclass
class BaseOutputs(BaseVarInfos):
    pass



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import sys
from typing import Optional

import numpy as np
import onnxruntime
import pytest

from spox import _debug, _value_prop
from spox._future import set_type_warning_level
from spox._graph import Graph
from spox._node import TypeWarningLevel

set_type_warning_level(TypeWarningLevel.CRITICAL)
_value_prop.VALUE_PROP_STRICT_CHECK = True
_debug.STORE_TRACEBACK = True


class ONNXRuntimeHelper:
    _build_cache: dict[Graph, bytes]
    _last_graph: Optional[Graph]
    _last_session: Optional[onnxruntime.InferenceSession]

    def __init__(self):
        self._build_cache = {}
        self._last_graph = None
        self._last_session = None

    def run(self, graph: Graph, unwrap: Optional[str] = None, **kwargs):
        debug_index = {
            **graph._get_build_result().scope.var.of_name,
            **graph._get_build_result().scope.node.of_name,
        }
        if self._last_graph is graph:
            print("[ONNXRuntimeHelper] Reusing previous session.", file=sys.stderr)
            session = self._last_session
        else:
            if graph not in self._build_cache:
                model_proto = graph.to_onnx_model()
                model_bytes = model_proto.SerializeToString()
                self._build_cache[graph] = model_bytes
            else:
                model_bytes = self._build_cache[graph]
            with _debug.show_construction_tracebacks(debug_index):
                session = onnxruntime.InferenceSession(model_bytes)
        self._last_graph = graph
        self._last_session = session
        assert isinstance(session, onnxruntime.InferenceSession)
        with _debug.show_construction_tracebacks(debug_index):
            result = {
                output.name: result
                for output, result in zip(
                    session.get_outputs(), session.run(None, kwargs)
                )
            }
        if unwrap is not None:
            return result[unwrap]
        return result

    @staticmethod
    def assert_close(given, expected, rtol=1e-7):
        if given is None:
            assert expected is None
        else:
            if isinstance(given, list):
                for subarray in given:
                    np.testing.assert_allclose(
                        given, np.array(expected, dtype=subarray.dtype), rtol=rtol
                    )
            else:
                np.testing.assert_allclose(
                    given, np.array(expected, dtype=given.dtype), rtol=rtol
                )


@pytest.fixture(scope="session")
def onnx_helper():
    return ONNXRuntimeHelper()



# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

# ruff: noqa: E741 -- Allow ambiguous variable name
from collections.abc import Iterable
from dataclasses import dataclass
from typing import (
    Optional,
)

import numpy as np

from spox._attributes import (
    AttrFloat32,
    AttrFloat32s,
    AttrInt64,
    AttrInt64s,
    AttrString,
    AttrStrings,
    AttrTensor,
)
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._node import OpType
from spox._standard import StandardNode
from spox._var import Var, _VarInfo, create_prop_dict, unwrap_vars
from spox.opset.ai.onnx.ml.v3 import (
    _ArrayFeatureExtractor,
    _Binarizer,
    _CastMap,
    _CategoryMapper,
    _DictVectorizer,
    _FeatureVectorizer,
    _Imputer,
    _LinearClassifier,
    _LinearRegressor,
    _Normalizer,
    _OneHotEncoder,
    _Scaler,
    _SVMClassifier,
    _SVMRegressor,
    _TreeEnsembleClassifier,
    _TreeEnsembleRegressor,
    _ZipMap,
    array_feature_extractor,
    binarizer,
    cast_map,
    category_mapper,
    dict_vectorizer,
    feature_vectorizer,
    imputer,
    linear_classifier,
    linear_regressor,
    normalizer,
    one_hot_encoder,
    scaler,
    svmclassifier,
    svmregressor,
    tree_ensemble_classifier,
    tree_ensemble_regressor,
    zip_map,
)


class _LabelEncoder(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        default_float: AttrFloat32
        default_int64: AttrInt64
        default_string: AttrString
        default_tensor: Optional[AttrTensor]
        keys_floats: Optional[AttrFloat32s]
        keys_int64s: Optional[AttrInt64s]
        keys_strings: Optional[AttrStrings]
        keys_tensor: Optional[AttrTensor]
        values_floats: Optional[AttrFloat32s]
        values_int64s: Optional[AttrInt64s]
        values_strings: Optional[AttrStrings]
        values_tensor: Optional[AttrTensor]

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    op_type = OpType("LabelEncoder", "ai.onnx.ml", 4)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


def label_encoder(
    X: Var,
    *,
    default_float: float = -0.0,
    default_int64: int = -1,
    default_string: str = "_Unused",
    default_tensor: Optional[np.ndarray] = None,
    keys_floats: Optional[Iterable[float]] = None,
    keys_int64s: Optional[Iterable[int]] = None,
    keys_strings: Optional[Iterable[str]] = None,
    keys_tensor: Optional[np.ndarray] = None,
    values_floats: Optional[Iterable[float]] = None,
    values_int64s: Optional[Iterable[int]] = None,
    values_strings: Optional[Iterable[str]] = None,
    values_tensor: Optional[np.ndarray] = None,
) -> Var:
    r"""
    Maps each element in the input tensor to another value. The mapping is
    determined by the two parallel attributes, 'keys\_\ *' and 'values\_*'
    attribute. The i-th value in the specified 'keys\_\ *' attribute would
    be mapped to the i-th value in the specified 'values\_*' attribute. It
    implies that input's element type and the element type of the specified
    'keys\_\ *' should be identical while the output type is identical to
    the specified 'values\_*' attribute. Note that the 'keys\_\ *' and
    'values\_*' attributes must have the same length. If an input element
    can not be found in the specified 'keys\_\ *' attribute, the
    'default\_*' that matches the specified 'values\_\ *' attribute may be
    used as its output value. The type of the 'default\_*' attribute must
    match the 'values\_\ *' attribute chosen. Let's consider an example
    which maps a string tensor to an integer tensor. Assume and
    'keys_strings' is ["Amy", "Sally"], 'values_int64s' is [5, 6], and
    'default_int64' is '-1'. The input ["Dori", "Amy", "Amy", "Sally",
    "Sally"] would be mapped to [-1, 5, 5, 6, 6]. Since this operator is an
    one-to-one mapping, its input and output shapes are the same. Notice
    that only one of 'keys\_*'/'values\_\*' can be set. Float keys with
    value 'NaN' match any input 'NaN' value regardless of bit value. If a
    key is repeated, the last key takes precedence.

    Parameters
    ==========
    X
        Type T1.
        Input data. It must have the same element type as the keys\_\* attribute
        set.
    default_float
        Attribute.
        A float.
    default_int64
        Attribute.
        An integer.
    default_string
        Attribute.
        A string.
    default_tensor
        Attribute.
        A default tensor. {"*Unused"} if values*\ \* has string type, {-1} if
        values\_\* has integral type, and {-0.f} if values\_\* has float type.
    keys_floats
        Attribute.
        A list of floats.
    keys_int64s
        Attribute.
        A list of ints.
    keys_strings
        Attribute.
        A list of strings.
    keys_tensor
        Attribute.
        Keys encoded as a 1D tensor. One and only one of 'keys\_\*'s should be
        set.
    values_floats
        Attribute.
        A list of floats.
    values_int64s
        Attribute.
        A list of ints.
    values_strings
        Attribute.
        A list of strings.
    values_tensor
        Attribute.
        Values encoded as a 1D tensor. One and only one of 'values\_\*'s should
        be set.

    Returns
    =======
    Y : Var
        Type T2.
        Output data. This tensor's element type is based on the values\_\*
        attribute set.

    Notes
    =====
    Signature: ``ai.onnx.ml@4::LabelEncoder``.

    Type constraints:
     - T1: `tensor(double)`, `tensor(float)`, `tensor(int16)`, `tensor(int32)`, `tensor(int64)`, `tensor(string)`
     - T2: `tensor(double)`, `tensor(float)`, `tensor(int16)`, `tensor(int32)`, `tensor(int64)`, `tensor(string)`
    """
    input_prop_values = create_prop_dict(
        X=X,
    )
    output_vars = (
        _LabelEncoder(
            _LabelEncoder.Attributes(
                default_float=AttrFloat32(default_float, name="default_float"),
                default_int64=AttrInt64(default_int64, name="default_int64"),
                default_string=AttrString(default_string, name="default_string"),
                default_tensor=AttrTensor.maybe(default_tensor, name="default_tensor"),
                keys_floats=AttrFloat32s.maybe(keys_floats, name="keys_floats"),
                keys_int64s=AttrInt64s.maybe(keys_int64s, name="keys_int64s"),
                keys_strings=AttrStrings.maybe(keys_strings, name="keys_strings"),
                keys_tensor=AttrTensor.maybe(keys_tensor, name="keys_tensor"),
                values_floats=AttrFloat32s.maybe(values_floats, name="values_floats"),
                values_int64s=AttrInt64s.maybe(values_int64s, name="values_int64s"),
                values_strings=AttrStrings.maybe(values_strings, name="values_strings"),
                values_tensor=AttrTensor.maybe(values_tensor, name="values_tensor"),
            ),
            _LabelEncoder.Inputs(
                X=unwrap_vars(X),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Y
    )
    return output_vars  # type: ignore


_OPERATORS = {
    "ArrayFeatureExtractor": _ArrayFeatureExtractor,
    "Binarizer": _Binarizer,
    "CastMap": _CastMap,
    "CategoryMapper": _CategoryMapper,
    "DictVectorizer": _DictVectorizer,
    "FeatureVectorizer": _FeatureVectorizer,
    "Imputer": _Imputer,
    "LabelEncoder": _LabelEncoder,
    "LinearClassifier": _LinearClassifier,
    "LinearRegressor": _LinearRegressor,
    "Normalizer": _Normalizer,
    "OneHotEncoder": _OneHotEncoder,
    "SVMClassifier": _SVMClassifier,
    "SVMRegressor": _SVMRegressor,
    "Scaler": _Scaler,
    "TreeEnsembleClassifier": _TreeEnsembleClassifier,
    "TreeEnsembleRegressor": _TreeEnsembleRegressor,
    "ZipMap": _ZipMap,
}

_CONSTRUCTORS = {
    "ArrayFeatureExtractor": array_feature_extractor,
    "Binarizer": binarizer,
    "CastMap": cast_map,
    "CategoryMapper": category_mapper,
    "DictVectorizer": dict_vectorizer,
    "FeatureVectorizer": feature_vectorizer,
    "Imputer": imputer,
    "LabelEncoder": label_encoder,
    "LinearClassifier": linear_classifier,
    "LinearRegressor": linear_regressor,
    "Normalizer": normalizer,
    "OneHotEncoder": one_hot_encoder,
    "SVMClassifier": svmclassifier,
    "SVMRegressor": svmregressor,
    "Scaler": scaler,
    "TreeEnsembleClassifier": tree_ensemble_classifier,
    "TreeEnsembleRegressor": tree_ensemble_regressor,
    "ZipMap": zip_map,
}

__all__ = [fun.__name__ for fun in _CONSTRUCTORS.values()]



# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from collections.abc import Hashable
from typing import Generic, TypeVar, overload

from ._node import Node
from ._var import _VarInfo

H = TypeVar("H", bound=Hashable)


class ScopeError(Exception):
    """Represents an error related to mishandling a scope, like overwriting or accessing a missing entry."""

    pass


class ScopeSpace(Generic[H]):
    """
    Represents the namespace of a scope for some type H, like ``Node`` or ``_VarInfo``.

    Methods (and operators) on the namespace work both ways: both with names (str) and the named type (H).
    So ``__getitem__`` (``ScopeSpace[item]``) may be used for both the name of an object and the object of a name.
    """

    name_of: dict[H, str]
    of_name: dict[str, H]
    reserved: set[str]
    base_name_counters: dict[str, int]
    parent: ScopeSpace[H] | None

    def __init__(
        self,
        name_of: dict[H, str] | None = None,
        of_name: dict[str, H] | None = None,
        reserved: set[str] | None = None,
        parent: ScopeSpace[H] | None = None,
    ):
        """
        Parameters
        ----------
        name_of
            Name of a given object in this namespace.
        of_name
            Object with a given name in this namespace.
        reserved
            Set of reserved names, taken up by anonymous objects.
        parent
            Namespace of a parent scope. Is accessed first before all checks, but never modified.
        """
        self.name_of = name_of.copy() if name_of is not None else {}
        self.of_name = of_name.copy() if of_name is not None else {}
        self.reserved = reserved.copy() if reserved is not None else set()
        self.parent = parent
        # Reference to a single `base_name_counters` object across
        # all scopes.
        # While the standard is more lenient in this respect, we
        # simply don't allow any name reuse across the entire model.
        self.base_name_counters = (
            parent.base_name_counters if parent is not None else dict()
        )

    def __contains__(self, item: str | H) -> bool:
        """Checks if a given name or object is declared in this (or outer) namespace."""
        return (
            (self.parent is not None and item in self.parent)
            or item in self.reserved
            or item in self.name_of
            or item in self.of_name
        )

    @overload
    def __getitem__(self, item: H) -> str: ...

    @overload
    def __getitem__(self, item: str) -> H: ...

    def __getitem__(self, item: str | H) -> str | H:
        """Access the name of an object or an object with a given name in this (or outer) namespace."""
        if self.parent is not None and item in self.parent:
            return self.parent[item]
        elif isinstance(item, str):
            return self.of_name[item]
        else:
            return self.name_of[item]

    @overload
    def __setitem__(self, key: str, value: H) -> None: ...

    @overload
    def __setitem__(self, key: H, value: str) -> None: ...

    def __setitem__(self, _key: str | H, _value: H | str) -> None:
        """Set the name of an object in exactly this namespace. Both ``[name] = obj`` and ``[obj] = name`` work."""
        if isinstance(_value, str):
            _key, _value = _value, _key
        assert isinstance(_key, str)
        key: str = _key
        value: H = _value  # type: ignore
        if key in self and self[key] != value:
            raise ScopeError(
                f"Failed to name {value}, as its name {key} "
                f"was already taken by {self[key]}."
            )
        if value in self:
            if key != self[value]:
                raise ScopeError(
                    f"Attempt to implicitly rename {value} to {key} from {self[value]}."
                )
            return
        self.of_name[key] = value
        self.name_of[value] = key

    def __delitem__(self, item: str | H) -> None:
        """Delete a both the name and object from exactly this namespace."""
        if isinstance(item, str):
            key, value = item, self.of_name[item]
        else:
            key, value = self.name_of[item], item
        if key not in self.of_name:
            raise ScopeError(f"Cannot remove missing key: {key}")
        if value not in self.name_of:
            raise ScopeError(f"Cannot remove missing value: {value}")
        del self.of_name[key]
        del self.name_of[value]

    def enum(self, base: str) -> str:
        """Find an unused name by enumerating the pattern ``f"{base}_{i}"`` through `i = 0, 1, ...`"""
        self.base_name_counters.setdefault(base, 0)

        name = f"{base}_{self.base_name_counters[base]}"
        self.base_name_counters[base] = self.base_name_counters[base] + 1
        return name

    def maybe_enum(self, base: str) -> str:
        """Attempt to use ``base`` as a name, or return the result of ``self.enum`` for it otherwise."""
        if base not in self.base_name_counters:
            self.base_name_counters[base] = 0
            return base
        return self.enum(base)

    def reserve(self, name: str) -> str:
        if name in self:
            raise ScopeError(f"Reserved name is already in use: {name}")
        self.reserved.add(name)
        return name


class Scope:
    """
    Class representing the state of an ONNX-rules scope.

    Has namespaces (represented by a ScopeSpace) for VarInfos and Nodes.
    """

    var: ScopeSpace[_VarInfo]
    node: ScopeSpace[Node]

    def __init__(
        self,
        sub_var: ScopeSpace[_VarInfo] | None = None,
        sub_node: ScopeSpace[Node] | None = None,
        parent: Scope | None = None,
    ):
        self.var = sub_var if sub_var is not None else ScopeSpace()
        self.node = sub_node if sub_node is not None else ScopeSpace()
        if parent is not None:
            self.var = ScopeSpace(self.var.name_of, self.var.of_name, parent=parent.var)
            self.node = ScopeSpace(
                self.node.name_of, self.node.of_name, parent=parent.node
            )

    @classmethod
    def of(
        cls,
        *what: tuple[str, _VarInfo | Node] | tuple[_VarInfo | Node, str],
    ) -> Scope:
        """Convenience constructor for filling a Scope with known names."""
        scope = cls()
        for key, value in what:
            if not isinstance(key, str):
                key, value = value, key
            assert isinstance(key, str)
            if isinstance(value, _VarInfo):
                scope.var[key] = value
            elif isinstance(value, Node):
                scope.node[key] = value
            else:
                raise TypeError(f"Unknown value type for Scope.of: {type(value)}")
        return scope

    def update(self, node: Node, prefix: str = "", force: bool = True) -> None:
        """
        Function used for introducing a Node and its outputs into the scope in the build routine.

        The node is named by the pattern ``{node.op_type.identifier}_{i}``, where ``i` is a generated index.

        The var is named by the pattern ``{node_name}_{output_field_name}``, unless it has a ``._name`` field set.
        (as arguments and results of the main graph do, for example).

        Parameters
        ----------
        node
            Node to introduce in the scope.
        prefix
            What value to prefix the node name with. If the VarInfo has a predeclared name, it does not get the prefix.
        force
            Whether to attempt to overwrite existing names (possibly raising a ScopeError if they were different).
            By default, this is set to True to be more strict, so we see if the scoping algorithm failed to only
            introduce a node once where it is needed.
        """
        if force or node not in self.node:
            self.node[node] = self.node.enum(prefix + node.op_type.identifier)
        for field, arr in node.outputs.get_var_infos().items():
            if arr._name is None:
                base = f"{self.node[node]}_{field}"
                name = self.var.maybe_enum(base)
            else:
                name = arr._name
            if force or arr not in self.var:
                self.var[arr] = name



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import itertools
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Callable, Optional

import onnx

from spox._exceptions import BuildError
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._internal_op import INTERNAL_MIN_OPSET, _InternalNode
from spox._node import OpType
from spox._scope import Scope
from spox._type_system import Type
from spox._var import _VarInfo

from . import _value_prop


def rename_in_graph(
    graph_: onnx.GraphProto,
    rename: Callable[[str], str],
    *,
    rename_node: Optional[Callable[[str], str]] = None,
    rename_op: Optional[Callable[[str, str], tuple[str, str]]] = None,
) -> onnx.GraphProto:
    def rename_in_subgraph(subgraph: onnx.GraphProto) -> onnx.GraphProto:
        return rename_in_graph(
            subgraph,
            rename,
            rename_node=rename_node,
            rename_op=rename_op,
        )

    graph = onnx.GraphProto()
    graph.CopyFrom(graph_)

    for p in itertools.chain(graph.input, graph.initializer):
        p.name = rename(p.name)
    for si in graph.sparse_initializer:
        si.values.name = rename(si.values.name)
        si.indices.name = rename(si.indices.name)

    for nd in graph.node:
        if nd.name and rename_node is not None:
            nd.name = rename_node(nd.name)
        if rename_op is not None:
            # This is a bit elaborate, but we do it this way as
            # an unset domain field is different from an empty one.
            if nd.HasField("domain"):
                nd.domain, nd.op_type = rename_op(nd.domain, nd.op_type)
            else:
                # An empty domain is the default domain (ai.onnx)
                domain, nd.op_type = rename_op("", nd.op_type)
                if domain:  # Only set the domain explicitly if it's changing
                    nd.domain = domain
        for seq in (nd.input, nd.output):
            for i, name in enumerate(seq):
                seq[i] = rename(name)
        for attr_proto in nd.attribute:
            attr = onnx.helper.get_attribute_value(attr_proto)
            if isinstance(attr, onnx.GraphProto):
                attr_proto.g.CopyFrom(rename_in_subgraph(attr))
            elif isinstance(attr, list) and all(
                isinstance(g, onnx.GraphProto) for g in attr
            ):
                for i, sub in enumerate(attr):
                    attr_proto.graphs[i].CopyFrom(rename_in_subgraph(sub))

    for p in itertools.chain(graph.output, graph.value_info):
        p.name = rename(p.name)

    return graph


class _Inline(_InternalNode):
    """Internal operator used for inlining (embedding) an existing ONNX ModelProto inside a Spox graph."""

    model: onnx.ModelProto

    @dataclass
    class Attributes(BaseAttributes):
        pass

    @dataclass
    class Inputs(BaseInputs):
        inputs: Sequence[_VarInfo]

    @dataclass
    class Outputs(BaseOutputs):
        outputs: Sequence[_VarInfo]

    op_type = OpType("Inline", "spox.internal", 0)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs

    def pre_init(self, **kwargs: Any) -> None:
        self.model = kwargs["model"]

    @property
    def graph(self) -> onnx.GraphProto:
        return self.model.graph

    @property
    def opset_req(self) -> set[tuple[str, int]]:
        return {(imp.domain, imp.version) for imp in self.model.opset_import} | {
            ("", INTERNAL_MIN_OPSET)
        }

    def infer_output_types(
        self, input_prop_values: _value_prop.PropDict
    ) -> dict[str, Type]:
        # First, type check that we match the ModelProto type requirements
        for i, var in zip(self.graph.input, self.inputs.inputs):
            if var.type is not None and not (
                var.type._subtype(Type._from_onnx(i.type))
            ):
                raise TypeError(
                    f"Input '{i.name}' to inlined model got type {var.type}, "
                    f"expected {Type._from_onnx(i.type)}."
                )
        # If we do, take the types as declared in the model
        return {
            f"outputs_{k}": Type._from_onnx(o.type)
            for k, o in enumerate(self.graph.output)
        }

    def propagate_values(
        self, input_prop_values: _value_prop.PropDict
    ) -> dict[str, _value_prop.PropValueType]:
        if any(
            var_info.type is None or input_prop_values.get(var_info.name) is None
            for var_info in self.model.graph.input
        ):
            return {}
        wrap_feed, run, unwrap_feed = _value_prop.get_backend_calls()
        input_feed = {
            i.name: wrap_feed(input_prop_values.get(i.name))
            for i in self.model.graph.input
        }
        output_feed = run(self.model, input_feed)
        return {
            f"outputs_{k}": unwrap_feed(var.unwrap_type(), output_feed[o.name]).value
            for k, (o, var) in enumerate(zip(self.graph.output, self.outputs.outputs))
            if o.name in output_feed
        }

    def to_onnx(
        self,
        scope: Scope,
        doc_string: Optional[str] = None,
        build_subgraph: Optional[Callable] = None,
    ) -> list[onnx.NodeProto]:
        input_names: dict[str, int] = {
            p.name: i for i, p in enumerate(self.graph.input)
        }
        output_names: dict[str, int] = {
            p.name: i for i, p in enumerate(self.graph.output)
        }
        inner_renames: dict[str, str] = {}
        inner_node_renames: dict[str, str] = {}

        def reserve_prefixed(name: str) -> str:
            if not name:
                return name
            return scope.var.reserve(
                scope.var.maybe_enum(f"{scope.node[self]}__{name}")
            )

        def apply_rename(name: str) -> str:
            if name in input_names:
                return scope.var[self.inputs.inputs[input_names[name]]]
            if name in output_names:
                return scope.var[self.outputs.outputs[output_names[name]]]
            if name not in inner_renames:
                inner_renames[name] = reserve_prefixed(name)
            return inner_renames[name]

        def apply_node_rename(name: str) -> str:
            if name not in inner_node_renames:
                inner_node_renames[name] = reserve_prefixed(name)
            return inner_node_renames[name]

        graph = rename_in_graph(self.graph, apply_rename, rename_node=apply_node_rename)

        if graph.initializer:
            raise BuildError(
                "Inlined graph initializers should be handled beforehand and be removed from the graph."
            )
        nodes: list[onnx.NodeProto] = list(graph.node)
        return nodes



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
from dataclasses import dataclass

import numpy as np
import onnx
import onnx.parser
import onnx.shape_inference
import onnxruntime
import pytest

import spox.opset.ai.onnx.v17 as op
from spox._attributes import Attr, AttrFloat32, _Ref
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs, BaseVars
from spox._function import Function, to_function
from spox._graph import arguments, results
from spox._node import OpType
from spox._type_system import Tensor
from spox._var import Var, _VarInfo


@pytest.fixture
def linear():
    class LinearFunction(Function):
        @dataclass
        class Attributes(BaseAttributes):
            slope_outer: AttrFloat32
            shift_outer: AttrFloat32

        @dataclass
        class Inputs(BaseInputs):
            X: _VarInfo

        @dataclass
        class Outputs(BaseOutputs):
            Y: _VarInfo

        op_type = OpType("LinearFunction", "spox.test", 0)

        attrs: Attributes
        inputs: Inputs
        outputs: Outputs

        def constructor(self, attrs: dict[str, Attr], inputs: BaseVars) -> Outputs:
            # FIXME: At some point, attribute references should be properly type-hinted.
            a = op.constant(
                value_float=_Ref(
                    attrs["slope_outer"], outer_name="slope_outer", name="value_float"
                )  # type: ignore
            )
            b = op.constant(
                value_float=_Ref(
                    attrs["shift_outer"], outer_name="shift_outer", name="value_float"
                )  # type: ignore
            )
            x = inputs.X
            return self.Outputs(op.add(op.mul(a, x), b)._var_info)

    def linear_inner(x: Var, a: float | _Ref[float], b: float | _Ref[float]) -> Var:
        return (
            LinearFunction(
                LinearFunction.Attributes(
                    slope_outer=AttrFloat32(a, "slope_outer"),
                    shift_outer=AttrFloat32(b, "shift_outer"),
                ),
                LinearFunction.Inputs(x._var_info),
            )
            .get_output_vars(input_prop_values={"x": x._value})
            .Y
        )

    return linear_inner


@pytest.fixture
def linear2(linear):
    class LinearFunction2(Function):
        @dataclass
        class Attributes(BaseAttributes):
            slope1: AttrFloat32
            shift1: AttrFloat32

        @dataclass
        class Inputs(BaseInputs):
            X: _VarInfo

        @dataclass
        class Outputs(BaseOutputs):
            Y: _VarInfo

        op_type = OpType("LinearFunction2", "spox.test", 0)

        attrs: Attributes
        inputs: Inputs
        outputs: Outputs

        def constructor(self, attrs: dict[str, Attr], inputs: BaseVars) -> Outputs:
            return self.Outputs(
                linear(
                    inputs.X,
                    _Ref(attrs["slope1"], outer_name="slope1", name="slope_outer"),
                    _Ref(attrs["shift1"], outer_name="shift1", name="shift_outer"),
                )._var_info
            )

    def linear_inner(x: Var, a: float | _Ref[float], b: float | _Ref[float]) -> Var:
        return (
            LinearFunction2(
                LinearFunction2.Attributes(
                    slope1=AttrFloat32(a, name="slope1"),
                    shift1=AttrFloat32(b, "shift1"),
                ),
                LinearFunction2.Inputs(x._var_info),
            )
            .get_output_vars({"X": x._value})
            .Y
        )

    return linear_inner


@pytest.fixture
def cubic(linear):
    class CubicFunction(Function):
        @dataclass
        class Attributes(BaseAttributes):
            a3: AttrFloat32
            a2: AttrFloat32
            a1: AttrFloat32
            a0: AttrFloat32

        @dataclass
        class Inputs(BaseInputs):
            X: _VarInfo

        @dataclass
        class Outputs(BaseOutputs):
            Y: _VarInfo

        op_type = OpType("CubicFunction", "spox.test.extra", 0)

        attrs: Attributes
        inputs: Inputs
        outputs: Outputs

        def constructor(self, attrs: dict[str, Attr], inputs: BaseVars) -> Outputs:
            x = inputs.X
            a = op.mul(
                linear(
                    x,
                    _Ref(attrs["a3"], outer_name="a3", name="slope_outer"),
                    _Ref(attrs["a2"], outer_name="a2", name="shift_outer"),
                ),
                op.mul(x, x),
            )
            b = op.add(
                op.mul(
                    x,
                    op.constant(
                        value_float=_Ref(
                            attrs["a1"], outer_name="a1", name="value_float"
                        )  # type: ignore
                    ),
                ),
                op.constant(
                    value_float=_Ref(attrs["a0"], outer_name="a0", name="value_float"),  # type: ignore
                ),
            )
            y = op.add(a, b)
            return self.Outputs(y._var_info)

    def cubic_inner(x: Var, a3: float, a2: float, a1: float, a0: float) -> Var:
        return (
            CubicFunction(
                CubicFunction.Attributes(
                    a3=AttrFloat32(a3, name="a3"),
                    a2=AttrFloat32(a2, name="a2"),
                    a1=AttrFloat32(a1, name="a1"),
                    a0=AttrFloat32(a0, name="a0"),
                ),
                CubicFunction.Inputs(X=x._var_info),
            )
            .get_output_vars()
            .Y
        )

    return cubic_inner


@pytest.fixture
def min_fun_graph(linear):
    (start,) = arguments(start=Tensor(np.float32, (None,)))
    return results(final=linear(start, 3, 2))


@pytest.fixture
def big_fun_graph(linear):
    first, second = arguments(
        first=Tensor(np.float32, (None,)), second=Tensor(np.float32, (None,))
    )
    return results(final=op.div(linear(op.add(first, second), 3, 2), second))


@pytest.fixture
def double_fun_graph(linear):
    (start,) = arguments(start=Tensor(np.float32, (None,)))
    return results(final=linear(linear(start, 3, 2), 5, 3))


@pytest.fixture
def wrapped_linear_graph(linear2):
    (start,) = arguments(start=Tensor(np.float32, (None,)))
    return results(final=linear2(start, 3, 2))


@pytest.fixture
def cubic_graph(cubic):
    (x,) = arguments(x=Tensor(np.float32, (None,)))
    return results(y=cubic(x, 5, 3, 2, 1))


@pytest.fixture
def cubic_rational_graph(cubic):
    (x,) = arguments(x=Tensor(np.float32, (None,)))
    return results(y=op.div(cubic(x, 5, 3, 2, 1), cubic(x, 3, 4, 2, 3)))


@pytest.fixture
def cubic_rational_graph_2x3(linear, cubic):
    (x,) = arguments(x=Tensor(np.float32, (None,)))
    return results(y=linear(op.div(cubic(x, 5, 3, 2, 1), cubic(x, 3, 4, 2, 3)), 2, 3))


@pytest.fixture
def isnan_graph():
    x, y, z = arguments(
        x=Tensor(np.float64, ()),
        y=Tensor(np.float64, ()),
        z=Tensor(np.float64, ()),
    )

    @to_function("IsNaN", "spox.test")
    def isnan(v: Var) -> list[Var]:
        return [op.not_(op.equal(v, v))]

    return results(
        count=functools.reduce(
            op.add, (op.cast(*isnan(w), to=np.int64) for w in (x, y, z))
        )
    )


def test_minimal(onnx_helper, min_fun_graph):
    a = np.random.rand(8).astype(np.float32)
    onnx_helper.assert_close(
        onnx_helper.run(min_fun_graph, "final", start=a), 3 * a + 2
    )


def test_bigger(onnx_helper, big_fun_graph):
    a = np.random.rand(8).astype(np.float32)
    b = np.random.rand(8).astype(np.float32) + 1
    onnx_helper.assert_close(
        onnx_helper.run(big_fun_graph, "final", first=a, second=b),
        (3 * (a + b) + 2) / b,
    )


def test_double_application(onnx_helper, double_fun_graph):
    a = np.random.rand(8).astype(np.float32)
    onnx_helper.assert_close(
        onnx_helper.run(double_fun_graph, "final", start=a), 5 * (3 * a + 2) + 3
    )


def test_onnxruntime_simple_function_support(onnx_helper):
    # See https://github.com/microsoft/onnxruntime/issues/10250 (fixed)
    import onnx.parser

    m = onnx.parser.parse_model(
        """
    <
      ir_version: 8,
      opset_import: [ "" : 14, "local" : 1],
      producer_name: "test",
      producer_version: "1.0",
      model_version: 1,
      doc_string: "Test preprocessing model"
    >
    agraph (uint8[H, W, C] x) => (uint8[H, W, C] x_processed)
    {
        x_processed = local.func(x)
    }

    <
      opset_import: [ "" : 14 ],
      domain: "local",
      doc_string: "function 1"
    >
    f1 (x) => (y) {
        y = Identity(x)
    }

    <
      opset_import: [ "" : 14 ],
      domain: "local",
      doc_string: "function 2"
    >
    f2 (x) => (y) {
        y = Identity(x)
    }

    <
      opset_import: [ "" : 14, "local" : 1 ],
      domain: "local",
      doc_string: "Preprocessing function."
    >
    func (x) => (y) {
        x1 = local.f1(x)
        y = local.f2(x1)
    }

    """
    )

    onnx.checker.check_model(m)
    onnxruntime.InferenceSession(m.SerializeToString())


@pytest.mark.xfail
def test_onnxruntime_nested_function_attr_support():
    m = onnx.parser.parse_model(
        """
    <
      ir_version: 8,
      opset_import: [ "" : 14, "local" : 1],
      model_version: 1
    >
    agraph (int64[H] x) => (int64[H] y)
    {
        y = local.func <outer = 1> (x)
    }

    <
      opset_import: [ "" : 14 ],
      domain: "local",
      doc_string: "function 1"
    >
    f1 <inner> (x) => (y) {
        c = Constant<value_int: int = @inner>()
        y = Add(x, c)
    }

    <
      opset_import: [ "" : 14, "local" : 1 ],
      domain: "local",
      doc_string: "Preprocessing function."
    >
    func <outer> (x) => (y) {
        y = local.f1 <inner: int = @outer> (x)
    }
    """
    )

    onnx.checker.check_model(m)
    session = onnxruntime.InferenceSession(m.SerializeToString())

    x = np.array([0, 0, 0], dtype=np.int64)
    y = np.array([1, 1, 1], dtype=np.int64)
    assert (session.run(None, {"x": x})[0] == y).all()


def test_minimal_wrapped(onnx_helper, wrapped_linear_graph):
    a = np.random.rand(8).astype(np.float32)
    onnx_helper.assert_close(
        onnx_helper.run(wrapped_linear_graph, "final", start=a), 3 * a + 2
    )


def test_simple_nested_calls_session(onnx_helper, cubic_graph):
    model = cubic_graph.to_onnx_model()
    onnxruntime.InferenceSession(model.SerializeToString())


def test_simple_nested_calls(onnx_helper, cubic_graph):
    a = np.random.rand(8).astype(np.float32)
    # increase rtol due to small *non-deterministic* discrepancies
    onnx_helper.assert_close(
        onnx_helper.run(cubic_graph, "y", x=a), (1 + 2 * a) + a * a * (3 + a * 5)
    )


@pytest.mark.skip(
    "ONNX Runtime generates colliding internal identifiers for nested function nodes."
)
def test_nested_calls(onnx_helper, cubic_rational_graph):
    a = np.random.rand(8).astype(np.float32)
    onnx_helper.assert_close(
        onnx_helper.run(cubic_rational_graph, "y", x=a),
        (1 + a * (2 + a * (3 + a * 5))) / (3 + a * (2 + a * (4 + a * 3))),
    )


@pytest.mark.skip(
    "ONNX Runtime generates colliding internal identifiers for function nodes."
)
def test_complex_nested_calls(onnx_helper, cubic_rational_graph_2x3):
    a = np.random.rand(8).astype(np.float32)
    onnx_helper.assert_close(
        onnx_helper.run(cubic_rational_graph_2x3, "y", x=a),
        2 * (1 + a * (2 + a * (3 + a * 5))) / (3 + a * (2 + a * (4 + a * 3))) + 3,
    )


def test_to_function_isnan(onnx_helper, isnan_graph):
    nan = np.array(np.nan, dtype=np.float64)
    onnx_helper.assert_close(
        onnx_helper.run(
            isnan_graph,
            "count",
            x=np.array(0.0),
            y=np.array(1.0),
            z=np.array(2.0),
        ),
        0,
    )
    onnx_helper.assert_close(
        onnx_helper.run(isnan_graph, "count", x=nan, y=np.array(3.0), z=nan),
        2,
    )
